@@ -500,7 +500,7 @@ int handle_sec_auth_session_close(int cfd, sec_mod_st *sec, const SecAuthSession
 		                (pack_func) cli_stats_msg__pack);
 	}
 
-	if (e->status != PS_AUTH_COMPLETED) {
+	if (e->status < PS_AUTH_COMPLETED) {
 		seclog(sec, LOG_DEBUG, "session close received in unauthenticated client %s "SESSION_STR"!", e->auth_info.username, e->auth_info.psid);
 		return send_msg(e, cfd, SM_CMD_AUTH_CLI_STATS, &rep,
 		                (pack_size_func) cli_stats_msg__get_packed_size,
@@ -522,6 +522,8 @@ int handle_sec_auth_session_close(int cfd, sec_mod_st *sec, const SecAuthSession
 	/* send reply */
 	rep.bytes_in = e->stats.bytes_in;
 	rep.bytes_out = e->stats.bytes_out;
+	rep.has_secmod_client_entries = 1;
+	rep.secmod_client_entries = sec_mod_client_db_elems(sec);
 
 	ret = send_msg(e, cfd, SM_CMD_AUTH_CLI_STATS, &rep,
 			(pack_size_func) cli_stats_msg__get_packed_size,
@@ -604,7 +606,7 @@ int handle_sec_auth_stats_cmd(sec_mod_st * sec, const CliStatsMsg * req)
 		e->stats.uptime = req->uptime;
 
 	if (req->has_invalidate_cookie && req->invalidate_cookie != 0) {
-		seclog(sec, LOG_INFO, "invalidating session user '%s' "SESSION_STR,
+		seclog(sec, LOG_INFO, "invalidating session of user '%s' "SESSION_STR,
 			e->auth_info.username, e->auth_info.psid);
 		e->status = PS_AUTH_USER_TERM;
 	}
